#!/usr/bin/env python2

#
#
# SourceForge(r) Enterprise Edition
# Copyright 2007-2014 CollabNet, Inc.  All rights reserved.
# http://www.collab.net
#

"""Sourceforge Subversion integration pre-commit script - given a repository
and a transaction id, verify that the commit is valid (for example, has
required associations specified)."""

import sys, os
import svn
import SourceForge
import SubversionUtil
from xml.sax import SAXParseException


class Error(Exception): pass

def doIt(pool, repository, txn, systemId):
    host = SourceForge.getRequired('sfmain.integration.listener_host')
    port = SourceForge.getRequired('sfmain.integration.listener_port')

    if systemId == None:
        systemId = SourceForge.getRequired('external_system_id')

    svnlook = SubversionUtil.createSVNLook(repository, txn = txn)
    logMsg = unicode(svnlook.log(), 'utf-8')
    author = svnlook.author()
    # log.write("commit message: %s\n" % logMsg)

    # On Windows, we DO NOT pass the actual repository's full path and instead
    # create a path based on the understanding that ALL Windows integration
    # servers will be configured with '/windows-scmroot' as the repository
    # root path since CTF doesn't work with Windows paths.
    if SourceForge.isWindows():
        old_repository = repository
        repository = SourceForge.getCTFWindowsRepositoryPath(repository)
        # log.write('Converting full Windows path to CTF full path: %s->%s' % (old_repository, repository))

    scm = SourceForge.getSOAPClient()
    key = SourceForge.createScmRequestKey()
    commitMessageResponseStr = scm.service.isValidCommitMessage(key, author, systemId, repository, logMsg)

    commitMessageResponseParts = commitMessageResponseStr.split('\n')
    curLine = commitMessageResponseParts.pop(0)

    # Subversion seems to only show output to standard error, and discards standard out
    print commitMessageResponseParts
    while len(commitMessageResponseParts) > 0:
        sys.stderr.write(commitMessageResponseParts.pop(0) + '\n')

    if curLine == 'false':
        return 1

    return 0

def main():
    args = sys.argv
    if len(args) < 3:
        print 'Error: not enough args:', args
        return 1

    repository = unicode(args[1], 'utf-8')
    txn = unicode(args[2], 'utf-8')

    if len(args) >= 4:
        systemId = args[3]
    else:
        systemId = None

    if not SubversionUtil.isSVNRepository(repository):
        raise Error, "path '%s' doesn't look like a repository" % repository

    # log.write('repository: %s\n' % repository)
    # log.write('txn: %s\n\n' % txn)
    return svn.core.run_app(doIt, repository, txn, systemId)


def logException(log, inst):
    import traceback
    # log.write('Failed in pre-commit: ' + `inst` + '\n')
    # traceback.print_tb(sys.exc_info()[2], None, log)
    # log.close()

if __name__ == "__main__":

    # log = SourceForge.getLogFile('pre-commit.log')
    # log.setLogging(False)
#    SOAPpy.Config.debug = 1
#    sys.stdout = log

    result = 0
    try:
        result = main()
    except SAXParseException, inst:
        sys.stderr.write("Error: XML parse exception: %s\n" % str(inst))
        # logException(log, inst)
        result = 1
    except Exception, inst:
        sys.stderr.write("%s: %s" % (inst.__class__, str(inst)))
        # logException(log, inst)
        result = 1

    # log.close()
    sys.exit(result)

