#!/usr/bin/env python
# -*-python-*-

import sys
import os
import string
import struct
import time
import ConfigParser
import urllib2
import json
import base64
from datetime import timedelta
from functools import update_wrapper
from flask import Flask, jsonify, abort, Response, make_response, \
     request, current_app, redirect
import svnapi
import jwt
import jwt.contrib.algorithms
import jwt.contrib.algorithms.pycrypto
from jwt.contrib.algorithms.pycrypto import RSAAlgorithm
from Crypto.PublicKey import RSA


# ------------------------------------------------------------------------
# Adapted from http://flask.pocoo.org/snippets/56/
#

def crossdomain(default_origin='*',
                methods=['OPTIONS', 'GET'],
                headers=['Authorization', 'Content-Type'],
                max_age=21600,
                attach_to_all=True,
                automatic_options=True):
  if methods is not None:
    methods = ', '.join(sorted(x.upper() for x in methods))
  if headers is not None and not isinstance(headers, basestring):
    headers = ', '.join(x.upper() for x in headers)
  if not isinstance(default_origin, basestring):
    default_origin = ', '.join(default_origin)
  if isinstance(max_age, timedelta):
    max_age = max_age.total_seconds()

  def get_methods():
    if methods is not None:
      return methods

    options_resp = current_app.make_default_options_response()
    return options_resp.headers['allow']

  def decorator(f):
    def wrapped_function(*args, **kwargs):

      if automatic_options and request.method == 'OPTIONS':
        resp = current_app.make_default_options_response()
      else:
        resp = make_response(f(*args, **kwargs))
      if not attach_to_all and request.method != 'OPTIONS':
        return resp

      h = resp.headers
      h['Access-Control-Allow-Origin'] = default_origin
      h['Access-Control-Allow-Credentials'] = 'true'
      h['Access-Control-Allow-Methods'] = get_methods()
      h['Access-Control-Max-Age'] = str(max_age)
      if headers is not None:
        h['Access-Control-Allow-Headers'] = headers
      return resp

    f.provide_automatic_options = False
    return update_wrapper(wrapped_function, f)
  return decorator


# ------------------------------------------------------------------------
# General Helper Functions
#

def logerr(message):
  request.environ.get('wsgi.errors', sys.stderr).write(
    "[svn-rest] %s\n" % (message))

def _boolparam(paramname):
  return request.args.get(paramname, 'false') == 'true'

def _intparam(paramname, default):
  try:
    return int(request.args.get(paramname, default))
  except:
    return default

def _utf8path(path):
  return path.encode('utf-8')

def config2dict(config_path):
  cfg = ConfigParser.ConfigParser()
  cfg.read(config_path)
  def _config_section_map(section):
    return dict(map(lambda x: (x, cfg.get(section, x)),
                    cfg.options(section)))
  return dict(map(lambda x: (x, _config_section_map(x)),
                  cfg.sections()))

def config_get(cfg, section, option, default):
  return cfg.get(section, {}).get(option, default)

# ------------------------------------------------------------------------
# Authentication Functions
# (Some code adapted from https://github.com/rohe/pyjwkest)
#

def base64url_decode(b64):
  b64_len = len(b64) % 4
  if b64_len:
    b64 += b"=" * (4 - b64_len)
  return base64.urlsafe_b64decode(b64.encode("utf-8"))

def base64_to_long(b64):
  b64 = b64.encode("utf-8")
  decoded = base64.urlsafe_b64decode(bytes(b64) + b'==')
  return long(''.join(["%02x" % b for b in \
                      struct.unpack('%sB' % len(decoded), decoded)]), 16)

def ensure_keychain():
  import SourceForge
  global oauth_rsa_keys

  # If we've cached a set of keys, return them.
  if oauth_rsa_keys:
    return oauth_rsa_keys

  # If we aren't already caching keys, fetch the keychain from the
  # application (allowing for a few retries if the fetch fails).  Once
  # we get the keychain, we'll create and keep RSA key objects for the
  # keys of interest therein.
  # New endpoint /oauth/auth/token/keyset is introduced in 17.1, but keeping
  # /sf/... until SvnEdge no longer supports TeamForge 16.x
  keychain_url = SourceForge.getAPIServerBaseUrl() + '/sf/auth/token/keyset'
  for delay in [0.5, 2.5, 5.0, None]:
    try:
      keychain = json.load(urllib2.urlopen(keychain_url, timeout=10))

      # Keychain looks something like this:
      #
      # { 'keys': [
      #   { 'use': 'sig',
      #     'e': 'AQAB',
      #     'kty': 'RSA',
      #     'alg': 'RS256',
      #     'n': '-AML1G...',
      #     'kid': '1'
      #   },
      #   ...
      #   ]
      # }
      for key in keychain['keys']:
        if (key['kty'] == 'RSA' and
            key['alg'] == 'RS256' and
            key.has_key('n') and
            key.has_key('e')):
          if key.has_key('d'):
            rsa_key = RSA.construct((base64_to_long(key['n']),
                                     base64_to_long(key['e']),
                                     base64_to_long(key['d'])))
          else:
            rsa_key = RSA.construct((base64_to_long(key['n']),
                                     base64_to_long(key['e'])))
          if oauth_rsa_keys is None:
            oauth_rsa_keys = {}
          oauth_rsa_keys[key['kid']] = rsa_key
      return oauth_rsa_keys
    except:
      if delay is not None:
        time.sleep(delay)

  # If we get here, we've failed to fetch and parse the keychain.
  logerr("Error parsing OAuth keyset")
  abort(500)

def verify_bearer(bearer):
  # Bearer is a dotted triplet of base64url-encoded bits.
  try:
    header, payload, signature = bearer.split('.', 2)
    header = json.loads(base64url_decode(header))
    payload = json.loads(base64url_decode(payload))
  except:
    logerr("Malformed bearer token")
    abort(500)

  # Get the OAuth keychain and specific key against which we validate.
  rsa_key = None
  try:
    rsa_key = ensure_keychain()[header['kid']]
  except:
    logerr("Requested OAuth key not found (kid=%s)" % (header['kid']))
    abort(500)

  # Try to decode and validate the bearer token.  If we succeed,
  # return the username.  Any failure falls through to a 401 abort.
  try:
    verify_options = {
      'verify_aud': True,
      'verify_exp': True,
      'verify_iat': True,
      'verify_nbf': True,
      'verify_signature': True,
      }
    return jwt.decode(bearer, rsa_key, audience='svn',
                      leeway=30, options=verify_options)['sub']
  except jwt.InvalidTokenError, e: # PyJWT errors subclass this
    logerr("Error decoding OAuth token: %s" % (str(e)))
  except Exception, e:
    logerr("Unknown error decoding OAuth token: %s" % (str(e)))
  abort(401)

def authorize_request():
  # Check for authorization details.  We check the following places, in order:
  # 1. The 'access_token' query parameter
  # 2. The request's Authorization header
  # 3. The HTTP_AUTHORIZATION environment variable[*]
  #
  # [*] mod_wsgi with "WSGIPassAuthorization on" uses this.
  username = None
  access_token = request.args.get('access_token', None)
  authorization = request.headers.get('Authorization',
                                      request.environ.get('HTTP_AUTHORIZATION'))
  if access_token:
    return verify_bearer(access_token)
  elif authorization:
    try:
      token_type, access_token = filter(None, authorization.split())
      token_type = token_type.lower()
      if token_type == 'bearer': # OAuth
        return verify_bearer(access_token)
      else:
        logerr("Unsupported authorization token type: %s" % (token_type))
        abort(401)
    except:
      logerr("Error parsing authorization token")
      abort(401)
  return 'nobody'

def api():
  # Determine the Subversion parent path.
  svn_parent_path = request.environ.get('SUBVERSION_PARENT_PATH',
                                        default_svn_parent_path)
  try:
    os.environ['SOURCEFORGE_HOME'] = request.environ['SOURCEFORGE_HOME']
  except:
    pass

  try:
    os.environ['SOURCEFORGE_PROPERTIES_PATH'] = request.environ['SOURCEFORGE_PROPERTIES_PATH']
  except:
    pass

  username = authorize_request()
  if not username:
    logerr("Unable to determine effective username")
    abort(401)
  return svnapi.SvnAPIHandler(svn_parent_path, username,
                              app.config.get('BYPASS_AUTHZ', False))

# ------------------------------------------------------------------------
# Core Application
#

# Initialize state.
default_svn_parent_path = '/var/svn'
debug_flag = True
oauth_rsa_keys = None # { kid => RSA_key }

# Fetch the configuration.
config_path = os.path.join(os.path.dirname(os.path.abspath(__file__)),
                           'apiservice.conf')
if os.path.exists(config_path):
  config = config2dict(config_path)
  default_svn_parent_path = config_get(config, 'general',
                                       'svn_parent_path', '/var/svn')
  debug_flag = config_get(config, 'general', 'debug', 'false') == 'true'

# Register a handler with JWT for the signature algorithm we use.  If
# it's already loaded, unload and reload it so we can be sure to use
# the version of the algorithm we're expecting.  (If the cryptography
# module is available, PyJWT will have already grabbed that module's
# algorithm.)
try:
  jwt.register_algorithm('RS256', RSAAlgorithm(RSAAlgorithm.SHA256))
except ValueError:
  jwt.unregister_algorithm('RS256')
  jwt.register_algorithm('RS256', RSAAlgorithm(RSAAlgorithm.SHA256))

# Create and configure the Flask app.
app = Flask(__name__)
app.debug = debug_flag


# ------------------------------------------------------------------------
# Custom Error Handlers
#

def make_error_response(message, statuscode):
  return Response("%s [status=%d]" % (message, statuscode),
                  status=statuscode, mimetype='text/plain')

@app.errorhandler(401)
@crossdomain()
def custom_401(error):
  return make_error_response(
    "Do I know you?", 401)

@app.errorhandler(403)
@crossdomain()
def custom_403(error):
  return make_error_response(
    "You shall not pass!", 403)

@app.errorhandler(404)
@crossdomain()
def custom_404(error):
  return make_error_response(
    "These aren't the droids you're looking for.", 404)

@app.errorhandler(500)
@crossdomain()
def custom_500(error):
  return make_error_response(
    "I solemnly swear that I am up to no good.", 500)


# ------------------------------------------------------------------------
# Routes
#

@app.route('/',
           methods=['OPTIONS', 'GET'])
@crossdomain()
def index():
  response = {
    "_type": "collection",
    "paged": False,
    "items": [
      {
        "_type": "version",
        "versionNumber": "1",
        "deprecated": False
      }],
    "_links": {
      "self": {"href": "/svn-rest"}
    }
  }
  return jsonify(response)


### FIXME(v2): This route should end with '/', but existing consumers
### transmit without.
@app.route('/1/revision/<repo_name>/<revision>',
           methods=['OPTIONS', 'GET'])
@crossdomain()
def revision(repo_name, revision=-1):
  response = api().revision({
    'repo_name': repo_name,
    'revision': revision,
    })
  return jsonify(response)


@app.route('/1/listdir/<repo_name>/<revision>/',
           methods=['OPTIONS', 'GET'])
@app.route('/1/listdir/<repo_name>/<revision>/<path:path>',
           methods=['OPTIONS', 'GET'])
@crossdomain()
def listdir(repo_name, revision=-1, path=""):
  repo_name = _utf8path(repo_name)
  path = _utf8path(path)
  response = api().listdir({
    'repo_name': repo_name,
    'revision': revision,
    'path': path,
    'collapse_empty_dirs': _boolparam('collapse_empty_dirs'),
    })
  return jsonify(response)


@app.route('/1/getfile/<repo_name>/<revision>/<path:path>',
           methods=['OPTIONS', 'GET'])
@crossdomain()
def getfile(repo_name, revision=-1, path=""):
  repo_name = _utf8path(repo_name)
  path = _utf8path(path)
  response = api().getfile({
    'repo_name': repo_name,
    'revision': revision,
    'path': path,
    })
  return jsonify(response)


@app.route('/1/getfileraw/<repo_name>/<revision>/<path:path>',
           methods=['OPTIONS', 'GET'])
@crossdomain()
def getfileraw(repo_name, revision=-1, path=""):
  headers = {}
  repo_name = _utf8path(repo_name)
  path = _utf8path(path)
  filename = os.path.basename(path)
  download = _boolparam('download')
  try:
    response = api().getfileraw({
      'repo_name': repo_name,
      'revision': revision,
      'path': path,
      'skip_encoding_detection': download,
      })
  except svnapi.SvnAPIPathNotFoundError:
    logerr("Path '%s' not found in revision '%s' of repository '%s'"
           % (path, revision, repo_name))
    abort(404)

  # Ensure that we have a MIME type.  Flask appears to use 'text/html'
  # by default, which isn't cool.
  mime_type = response.get('mimeType')
  if not mime_type:
    mime_type = 'application/octet-stream'

  # Send a Content-Length if we know it.
  file_size = response.get('fileSize')
  if file_size is not None:
    headers['Content-Length'] = file_size

  # Is this MIME type one that we allow to pass through the interface
  # unmodified?  (Unless this is a web-friend image type (GIF, JPEG,
  # PNG) or plain text with a specified encoding, the answer is "no".)
  is_okay_mimetype = ((mime_type in ['image/jpeg', 'image/gif', 'image/png']) or
                      (mime_type.startswith('text/plain; charset=')))

  # Fix up unacceptable mimetypes.
  if not is_okay_mimetype:
    mime_type = 'application/octet-stream'

  # If we're in download mode or had an unacceptable mimetype,
  # generate a header that essentially asks the User Agent to present
  # a "Save As..." dialog box.
  if download or not is_okay_mimetype:
    valid_chars = "-_.,() %s%s" % (string.ascii_letters, string.digits)
    filename = ''.join(c for c in filename if c in valid_chars)
    headers['Content-Disposition'] = 'attachment; filename="%s"' % (filename)

  # Make and return a Response object.
  return Response(response=response.get('fileContent'),
                  status=200, headers=headers, mimetype=mime_type)


@app.route('/1/getfilelines/<repo_name>/<revision>/<path:path>',
           methods=['OPTIONS', 'GET'])
@crossdomain()
def getfilelines(repo_name, revision=-1, path=""):
  repo_name = _utf8path(repo_name)
  path = _utf8path(path)
  response = api().getfilelines({
    'repo_name': repo_name,
    'revision': revision,
    'path': path,
    'start_line': _intparam('start_line', 0),
    'end_line': _intparam('end_line', -1),
    'allow_truncate': _boolparam('allow_truncate'),
    })
  return jsonify(response)


@app.route('/1/blame/<repo_name>/<revision>/<path:path>',
           methods=['OPTIONS', 'GET'])
@crossdomain()
def blame(repo_name, revision=-1, path=""):
  repo_name = _utf8path(repo_name)
  path = _utf8path(path)
  response = api().blame({
    'repo_name': repo_name,
    'revision': revision,
    'path': path,
    })
  return jsonify(response)


### FIXME(v2): This route should end with '/', but existing consumers
### transmit without.
@app.route('/1/getprops/<repo_name>/<revision>',
           methods=['OPTIONS', 'GET'])
@app.route('/1/getprops/<repo_name>/<revision>/<path:path>',
           methods=['OPTIONS', 'GET'])
@crossdomain()
def getprops(repo_name, revision=-1, path=""):
  repo_name = _utf8path(repo_name)
  path = _utf8path(path)
  response = api().getprops({
    'repo_name': repo_name,
    'revision': revision,
    'path': path,
    })
  return jsonify(response)


@app.route('/1/getlogs/<repo_name>/<revision>/',
           methods=['OPTIONS', 'GET'])
@app.route('/1/getlogs/<repo_name>/<revision>/<path:path>',
           methods=['OPTIONS', 'GET'])
@crossdomain()
def getlogs(repo_name, revision=-1, path=""):
  repo_name = _utf8path(repo_name)
  path = _utf8path(path)
  response = api().gethistory({
    'repo_name': repo_name,
    'revision': revision,
    'path': path,
    'numlogs': _intparam('numlogs', 20),
    })
  return jsonify(response)


@app.route('/1/lastdiff/<difftype>/<repo_name>/<revision>/',
           methods=['OPTIONS', 'GET'])
@app.route('/1/lastdiff/<difftype>/<repo_name>/<revision>/<path:path>',
           methods=['OPTIONS', 'GET'])
@crossdomain()
def lastdiff(difftype, repo_name, revision=-1, path=""):
  repo_name = _utf8path(repo_name)
  path = _utf8path(path)
  response = api().lastdiff({
    'repo_name': repo_name,
    'revision': revision,
    'path': path,
    'type': difftype,
    'context_lines': _intparam('context_lines', 5),
    'base_revision': _intparam('base_revision', -1),
    })
  return jsonify(response)


### FIXME(v2): This route should end with '/', but existing consumers
### transmit without.
@app.route('/1/pathatrev/<repo_name>/<revision>',
           methods=['OPTIONS', 'GET'])
@app.route('/1/pathatrev/<repo_name>/<revision>/<path:path>',
           methods=['OPTIONS', 'GET'])
@crossdomain()
def pathatrev(repo_name, revision=-1, path=""):
  repo_name = _utf8path(repo_name)
  path = _utf8path(path)
  response = api().pathatrev({
    'repo_name': repo_name,
    'revision': revision,
    'peg_revision' : _intparam('peg_revision', -1),
    'path': path,
    })
  return jsonify(response)


# ------------------------------------------------------------------------
# Documentation
#

@app.route('/spec',
           methods=['OPTIONS', 'GET'])
@crossdomain()
def spec():
  content = open(os.path.join(os.path.dirname(__file__),
                              'swagger.json'), 'r').read()
  return Response(response=content, status=200,
                  mimetype='application/json')


@app.route('/documentation',
           methods=['OPTIONS', 'GET'])
@crossdomain()
def documentation():
  import SourceForge
  swagger_url = SourceForge.getAPIServerBaseUrl() + '/apidoc/swagger'
  json_url = request.url_root + 'spec'
  return redirect(swagger_url + '?url=' + urllib2.quote(json_url),
                  code=302)


# ------------------------------------------------------------------------
# Main
#

if __name__ == '__main__':
  app.run(host=config_get(config, 'standalone', 'bind', '0.0.0.0'),
          port=int(config_get(config, 'standalone', 'port', '5000')),
          debug=True)
