# -*-python-*-
import sys
import os
import svn.repos, svn.fs, svn.delta, svn.client, svn.core
import mimetypes
import time
import urllib
import jsondiff
import ConfigParser
try:
  from cStringIO import StringIO
except:
  from StringIO import StringIO
mimetypes.init()

# Whether to enable logging of debug information (to
# /opt/collabnet/teamforge/log/httpd/svn-rest-debug.log).
ENABLE_LOGGING = False

# The maximum amount of a file's content that we will attempt to
# process (in bytes and megabytes).
MAX_FILE_LENGTH = (1024.0 * 1024.0 * 5.0)
MAX_FILE_LINES = 15000
MAX_FILE_LENGTH_MB = round(MAX_FILE_LENGTH / (1024.0 * 1024.0), 2)


#---------------------------------------------------------------------------

if ENABLE_LOGGING:
  def DEBUG_LOG(msg):
    fp = open('/opt/collabnet/teamforge/log/httpd/svn-rest-debug.log', 'a')
    fp.write(msg + '\n')
    fp.flush()
    fp.close()
else:
  def DEBUG_LOG(msg): pass


def _compare_paths(path1, path2):
  # Easy out: are the paths exactly the same?
  if path1 == path2:
    return 0

  # Setup for char-by-char comparison.
  path1_len = len(path1)
  path2_len = len(path2)
  min_len = min(path1_len, path2_len)
  i = 0

  # Skip past common prefix.
  while (i < min_len) and (path1[i] == path2[i]):
    i = i + 1

  # Get first differing characters.
  char1 = '\0'
  char2 = '\0'
  if (i < path1_len):
    char1 = path1[i]
  if (i < path2_len):
    char2 = path2[i]

  # Children of paths are greater than their parents, but less than
  # greater siblings of their parents.
  if (char1 == '/') and (i == path2_len):
    return 1
  if (char2 == '/') and (i == path1_len):
    return -1
  if (i < path1_len) and (char1 == '/'):
    return -1
  if (i < path2_len) and (char2 == '/'):
    return 1

  # Common prefix was skipped above, next character is compared to
  # determine order.
  return cmp(char1, char2)


def _properties2dict(props_path):
  """Parse a properties file into a dictionary mapping option names to
  values."""

  cfg = ConfigParser.ConfigParser()
  cfg.readfp(StringIO('[DUMMY]\n' + open(props_path, 'r').read()))
  return dict(map(lambda x: (x, cfg.get('DUMMY', x)), cfg.options('DUMMY')))


def _make_json_error(code, message, stacktrace=''):
  """Return a JSON object which carries error information (namely, a
  CODE and MESSAGE)."""

  response = {
    'errorCode': code,
    'errorMessage': message,
    }
  # Should JSON errors include stacktraces?  This is not acceptable for
  # production code, but can be helpful when debugging.
  if 0:
    response.update({
      'errorTraceback': stacktrace,
    })
  return response


def _get_mime_data(fsroot, path, guess_type=False):
  """Fetch the MIME type and character set stored as properties on
  PATH in FSROOT."""

  mime_type = None
  encoding = None
  svn_mime_type = svn.fs.node_proplist(fsroot, path).get(svn.core.SVN_PROP_MIME_TYPE)
  if svn_mime_type:
    mime_parts = map(lambda x: x.strip(), svn_mime_type.split(';'))
    type_subtype = mime_parts[0].lower()
    parameters = {}
    for part in mime_parts[1:]:
      name, value = part.split('=', 1)
      parameters[name] = value
    mime_type = type_subtype
    encoding = parameters.get('charset')
  if not mime_type and guess_type:
    mime_type = mimetypes.guess_type(path)[0]
  return mime_type, encoding


def _stacktrace():
  """Return the stack traceback associated with the current exception state."""

  exc_type, exc, exc_tb = sys.exc_info()
  try:
    import traceback
    tb = '\n'.join(traceback.format_exception(exc_type, exc, exc_tb))
  finally:
    del exc_tb
  return tb


def _svn2api_error(exc):
  """Transform a Subversion exception EXC into an API exception."""
  if exc.apr_err == svn.core.SVN_ERR_FS_NOT_FOUND:
    raise SvnAPIPathNotFoundError()
  if exc.apr_err == svn.core.SVN_ERR_CLIENT_IS_BINARY_FILE:
    raise SvnAPIAnnotationBinaryFileError()
  if exc.apr_err == svn.core.SVN_ERR_AUTHZ_UNREADABLE:
    raise SvnAPIPathNotFoundError()
  if exc.apr_err == svn.core.SVN_ERR_FS_NO_SUCH_REVISION:
    raise SvnAPIRevisionNotFoundError()
  raise SvnAPIGenericError("Something went wrong; Subversion error was "
                           "'%s'" % (exc.message))


def _json_error(exc):
  """Transform exception EXC into a JSON error representation."""

  try:
    # If this is an SvnAPIError, return its JSON representation.
    if isinstance(exc, SvnAPIError):
      return exc.to_json()

    # If this is a Subversion exception, re-raise it as an SvnAPIError.
    if isinstance(exc, svn.core.SubversionException):
      _svn2api_error(exc)

    # If we get here, this is some kind of exception we weren't
    # prepared for.
    raise SvnAPIGenericError("Something went wrong: %s" % (str(exc)))

  except SvnAPIError, e:
    # Something above was re-raised as an SvnAPIError.  Return the
    # JSON representation.
    return e.to_json()

  except:
    # Punt!
    return _make_json_error('FAILURE!',
                            'falla // Scheitern // xe5\xa4\xb1\xe8\xb4\xa5\x0a',
                            _stacktrace())


def _pop(d, key, default=None):
  if d.has_key(key):
    return d.pop(key)
  return default


class SvnContentProxy:
  def __init__(self, fsroot, path, max_size=MAX_FILE_LENGTH):
    self.stream = svn.fs.file_contents(fsroot, path)
    self.max_size = max_size
    self.processed = 0
    self.truncated = False

  def was_truncated(self):
    return self.truncated

  def get_content(self):
    if not self.stream:
      raise Exception("Content stream already closed")
    fp = StringIO()
    while 1:
      chunk = svn.core.svn_stream_read(self.stream,
                                       svn.core.SVN_STREAM_CHUNK_SIZE)
      if not chunk:
        break
      self.processed += len(chunk)
      if self.max_size and (self.processed > self.max_size):
        chunk = chunk[:(len(chunk) + self.max_size - self.processed)]
        self.truncated = True
      fp.write(chunk)
      if self.truncated:
        break
    self._close()
    filecontent = fp.getvalue()
    fp.close()
    return filecontent

  def get_generator(self):
    if not self.stream:
      raise Exception("Content stream already closed")
    while 1:
      chunk = svn.core.svn_stream_read(self.stream,
                                       svn.core.SVN_STREAM_CHUNK_SIZE)
      if not chunk:
        break
      self.processed += len(chunk)
      if self.max_size and (self.processed > self.max_size):
        chunk = chunk[:(len(chunk) + self.max_size - self.processed)]
        self.truncated = True
      yield chunk
      if self.truncated:
        break
    self._close()

  def get_lines(self, start_line=0, end_line=-1):
    if not self.stream:
      raise Exception("Content stream already closed")
    lines = []
    line_count = 0
    while 1:
      line, eof = svn.core.svn_stream_readline(self.stream, '\n')
      if (line is None) or (eof and not line):
        break
      self.processed += (len(line) + ((not eof) and 1 or 0))
      if self.max_size and (self.processed > self.max_size):
        line = line[:(len(line) + self.max_size - self.processed)]
        self.truncated = True
      if line_count < start_line:
        # Line preceeds our range of interest.  Ignore.
        pass
      elif end_line >= 0 and line_count > end_line:
        # Line succeeds our range of interest.  Close up show and go home.
        break
      else:
        lines.append(line)
      line_count += 1
      if eof or self.truncated:
        break
    self._close()
    return lines

  def _close(self):
    if self.stream:
      svn.core.svn_stream_close(self.stream)
      self.stream = None

  def __del__(self):
    self._close()


class SvnHistoryReceiver:
  """An iterable object that returns 2-tuples of (revision, path)
  locations along a node's change history, ordered from youngest to
  oldest."""

  def __init__(self, fs, limit=0):
    self.histories = []
    self.fs = fs
    self.oldest_rev = None
    self.limit = limit

  def add_history(self, path, revision, pool):
    if self.oldest_rev is None:
      self.oldest_rev = revision
    else:
      assert(revision < self.oldest_rev)

    self.histories.append([revision, path])
    if self.limit and len(self.histories) == self.limit:
      raise svn.core.SubversionException("", svn.core.SVN_ERR_CEASE_INVOCATION)

  def __getitem__(self, idx):
    return self.histories[idx]


class SvnAnnotator:
  def __init__(self, local_url, revision, start_line=0, end_line=-1):
    self.idx = -1
    self.blame_data = []
    self.revs = {}
    self.start_line = start_line;
    self.end_line = end_line;

    ctx = svn.client.create_context()
    ctx.config = {}
    ctx.auth_baton = svn.core.svn_auth_open([])
    try:
      svn.client.blame2(local_url, self._rev2optrev(revision),
                        self._rev2optrev(1), self._rev2optrev(revision),
                        self._blame_cb, ctx)
    except svn.core.SubversionException, e:
      if e.apr_err != svn.core.SVN_ERR_CEASE_INVOCATION:
        raise

  def _rev2optrev(self, revision):
    revision = int(revision)
    rt = svn.core.svn_opt_revision_t()
    rt.kind = svn.core.svn_opt_revision_number
    rt.value.number = revision
    return rt

  def _blame_cb(self, line_num, revision, author, date, text, pool):
    if line_num < self.start_line:
      return
    if self.end_line >= 0 and line_num > self.end_line:
      raise svn.core.SubversionException("", svn.core.SVN_ERR_CEASE_INVOCATION)
    self.revs[revision] = None
    self.blame_data.append({
      'lineNum': line_num + 1,
      'revMetaKey': str(revision),
      })

  def json_lines(self):
    return self.blame_data

  def revisions(self):
    return self.revs.keys()


class SvnDifferencer:
  def __init__(self, fsroot1, path1, fsroot2, path2,
               node_kind, unified=True, context_lines=3):
    self.fs = svn.fs.root_fs(fsroot1)
    self.fsroot1 = fsroot1
    self.path1 = path1
    self.fsroot2 = fsroot2
    self.path2 = path2
    self.node_kind = node_kind
    self.unified = unified
    self.context_lines = context_lines

  def _json_delta(self, lines1, lines2):
    delta = jsondiff.JsonDiff(lines1, lines2, self.context_lines)
    if self.unified:
      return delta.unified()
    else:
      return delta.sidebyside()

  def _string_delta(self, string1, string2):
    return self._json_delta(string1.splitlines(1), string2.splitlines(1))

  def _content_delta(self):
    import SourceForge
    if self.node_kind != svn.core.svn_node_file:
      return None
    filecontent1 = ""
    filecontent2 = ""
    if self.fsroot1 and self.path1:
      scp = SvnContentProxy(self.fsroot1, self.path1)
      filecontent1 = scp.get_content()
      mime_type, encoding = _get_mime_data(self.fsroot1, self.path1)
      if not encoding:
        encoding = SourceForge.detect_encoding(filecontent1)
      try:
        filecontent1 = SourceForge.transcode_to_utf8(filecontent1, encoding)
      except:
        raise SvnAPIUnsupportedContentEncodingError()
    if self.fsroot2 and self.path2:
      scp = SvnContentProxy(self.fsroot2, self.path2)
      filecontent2 = scp.get_content()
      mime_type, encoding = _get_mime_data(self.fsroot2, self.path2)
      if not encoding:
        encoding = SourceForge.detect_encoding(filecontent2)
      try:
        filecontent2 = SourceForge.transcode_to_utf8(filecontent2, encoding)
      except:
        raise SvnAPIUnsupportedContentEncodingError()
    delta = self._string_delta(filecontent2, filecontent1)
    return delta['diffLines'] and delta or None

  def _prop_deltas(self):
    props1 = {}
    props2 = {}
    if self.fsroot1 and self.path1 is not None:
      props1 = svn.fs.node_proplist(self.fsroot1, self.path1)
    if self.fsroot2 and self.path2 is not None:
      props2 = svn.fs.node_proplist(self.fsroot2, self.path2)
    prop_deltas = []
    for pname in props1.keys():
      if props2.has_key(pname):
        if props1[pname] != props2[pname]:
          val1 = props1[pname].encode('utf-8')
          val2 = props2[pname].encode('utf-8')
          delta = self._string_delta(val2, val1)
          delta['propName'] = pname
          prop_deltas.append(delta)
      else:
        val1 = props1[pname].encode('utf-8')
        delta = self._string_delta('', val1)
        delta['propName'] = pname
        prop_deltas.append(delta)
    for pname in props2.keys():
      if not props1.has_key(pname):
        val2 = props2[pname].encode('utf-8')
        delta = self._string_delta(val2, '')
        delta['propName'] = pname
        prop_deltas.append(delta)
    return prop_deltas

  def get_delta(self):
    return {
      'contentDelta': self._content_delta(),
      'propDeltas': self._prop_deltas(),
      }


class SvnAPIHandler:
  def __init__(self, svn_parent_path, username, bypass_authz=False):
    self.svn_parent_path = os.path.abspath(svn_parent_path)
    self.username = username

    DEBUG_LOG("Initializing SvnAPIHandler for user '%s'" % (username))

    # Mapping: rootname -> TFSvnAuthorizer
    self.auth = {}

    # A single-item array, where the item is a universal access
    # determination (1=yes, 0=no, None=dunno); or None if we haven't
    # queried for universal access yet.
    self.global_access = None

    # Allow the traversal of otherwise unreadable parent directories
    # of readable paths?
    self.readable_parentdirs = True

    # Timeout (in seconds) for checking access to a given file or
    # directory.
    self.path_authz_timeout = 10

    # Timeout (in seconds) for the harvesting of directory list
    # entries.
    self.listdir_timeout = 120

    # Do we honor the collapse_empty_dirs listdir() option?
    self.allow_collapse_empty_dirs = True

    if bypass_authz:
      self.global_access = [1]
    else:
      import SourceForge

      # Get some configuration values from the SourceForge module.
      self.readable_parentdirs = \
        SourceForge.getBoolean("svn.api.readable_parentdirs", "true")
      self.path_authz_timeout = \
        float(SourceForge.get("svn.api.path_authz_timeout", "10"))
      self.listdir_timeout = \
        float(SourceForge.get("svn.api.listdir_timeout", "120"))
      self.allow_collapse_empty_dirs = \
        SourceForge.getBoolean("svn.api.allow_collapse_empty_dirs", "true")

      # Read from the parent path the file that tells us the TeamForge
      # external system ID. Branding integration server does not have
      # .scm_properties file
      scm_properties_path = os.path.join(self.svn_parent_path,
                                         '.scm.properties')
      if os.path.exists(scm_properties_path):
        self.system_id = \
          _properties2dict(scm_properties_path).get('external_system_id')
      else:
        scm = SourceForge.getSOAPClient()
        key = SourceForge.createScmRequestKey()
        self.system_id = scm.service.getBrandingExternalSystemId(key)

  def _getauth(self, rootname):
    from ctfsvnauthz import TFSvnAuthorizer
    if not self.auth.has_key(rootname):
      DEBUG_LOG("Creating TFSvnAuthorizer instance for root '%s', "
                "system_id %s" % (rootname, self.system_id))
      rootpath = os.path.join(self.svn_parent_path, rootname)
      self.auth[rootname] = TFSvnAuthorizer(rootname, rootpath,
                                            self.username, self.system_id,
                                            self.readable_parentdirs)
      DEBUG_LOG("Completed TFSvnAuthorizer instantiation")
    return self.auth[rootname]

  def _has_universal_read_access(self, rootname):
    # If we don't already know the answer, ask (and remember).
    if self.global_access is None:
      DEBUG_LOG("Checking for universal access")
      auth = self._getauth(rootname)
      access = auth.determine_global_access()
      DEBUG_LOG("Completed check for universal access: %s"
                % (access or "(None)"))
      self.global_access = [access and 1 or None]

    # Return the access determination.
    return self.global_access[0]

  def _is_readable(self, rootname, path, rev, is_file):
    # Short circuit the authorization check when the user is known to
    # have a universal access determination (either granting or
    # denying access).
    univ_access = self._has_universal_read_access(rootname)
    if univ_access is not None:
      DEBUG_LOG("Checking access for %s '%s', r%d: %s universally"
                % (is_file and 'file' or 'directory', path, rev,
                   univ_access and "granted" or "DENIED"))
      return univ_access

    # No universal access grant.  We'll ask about this path/rev.
    # (Implied here is that this class was NOT instantiated with
    # bypass_authz=True.)
    from ctfsvnauthz import TFSvnAuthorizerTimeout
    if not path.startswith('/'):
      path = '/' + path
    try:
      auth = self._getauth(rootname)
      access = auth.check_path_access(path, rev, is_file,
                                      self.path_authz_timeout)
      DEBUG_LOG("Checking access for %s '%s', r%d: %s"
                % (is_file and 'file' or 'directory', path, rev,
                   access and "granted" or "DENIED"))
    except TFSvnAuthorizerTimeout:
      access = 0
      DEBUG_LOG("Checking access for %s '%s', r%d: DENIED [TIMEOUT]"
                % (is_file and 'file' or 'directory', path, rev))
    return access

  def _getpath(self, path):
    return svn.core.svn_path_canonicalize(path)

  def _getrepo(self, repo_name):
    try:
      return svn.repos.open(os.path.join(self.svn_parent_path, repo_name))
    except:
      raise SvnAPIPathNotFoundError()

  def _getroot(self, fs, revision):
    return svn.fs.revision_root(fs, revision)

  def _getrev(self, fs, revision):
    if revision == -1 or revision == "-1" or revision == "HEAD":
      return svn.fs.youngest_rev(fs)
    return int(revision)

  def _geturl(self, rootname, path):
    rootpath = os.path.join(self.svn_parent_path, rootname)
    drive, rootpath = os.path.splitdrive(rootpath)
    if os.sep != '/':
      rootpath = rootpath.replace(os.sep, '/')
    rootpath = urllib.quote(rootpath)
    path = urllib.quote(path)
    if drive:
      url = 'file:///' + drive + rootpath + '/' + path
    else:
      url = 'file://' + rootpath + '/' + path
    return svn.core.svn_path_canonicalize(url)

  def _filter_props(self, props):
    outprops = {}
    def _is_displayable(val):
      try:
        unicode(val, 'utf-8')
        return 1
      except:
        return 0
    for name, value in props.iteritems():
      if _is_displayable(name) and _is_displayable(value):
        outprops[name] = value
    return outprops

  def _node_kind_string(self, kind):
    if kind == svn.core.svn_node_file:
      return 'file'
    elif kind == svn.core.svn_node_dir:
      return 'dir'
    elif kind == svn.core.svn_node_none:
      return None
    else:
      raise Exception("Unsupported node kind")

  def _bypass_empty_tree(self, rootname, fsroot, path, child):
    this_path = path
    if child is not None:
      this_path = this_path + '/' + child
    entries = {}
    revision = svn.fs.revision_root_revision(fsroot)
    for somechild, somedirent in svn.fs.dir_entries(fsroot, this_path).items():
      # Authz check - keep only readable children
      if self._is_readable(rootname, this_path + '/' + somechild,
                           revision, somedirent.kind == svn.core.svn_node_file):
        entries[somechild] = somedirent
        # If we've found more than a single child that's readable,
        # there's no need to keep processing them.  This is not a
        # collapsible directory.
        if len(entries) > 1:
          break

    # If there's a single directory child in this directory, we'll
    # recurse to try to find the deepest such grandchild.
    if len(entries) == 1:
      for firstchild, firstdirent in entries.items():
        if firstdirent.kind == svn.core.svn_node_dir:
          DEBUG_LOG("Exploring intermediate directory '%s', r%d"
                    % (child + '/' + firstchild, revision))
          return self._bypass_empty_tree(rootname, fsroot, path,
                                         child + '/' + firstchild)
        break

    this_rev = svn.fs.node_created_rev(fsroot, this_path)
    this_crpath = svn.fs.node_created_path(fsroot, this_path)[1:]
    this_revmeta_key = str(this_rev)
    return {
      'name': os.path.basename(child),
      'parentTree': os.path.dirname(child) or None,
      'kind': self._node_kind_string(svn.core.svn_node_dir),
      'revMetaKey' : this_revmeta_key,
      'revPath': this_crpath,
      }

  def _get_revprops(self, rootname, repo, revision):
    univ_access = self._has_universal_read_access(rootname)
    proplist = None
    if univ_access:
      DEBUG_LOG("Fetching revision properties for r%d (authz bypassed)"
                % (revision))
      proplist = svn.repos.fs_revision_proplist(repo, revision, None)
    else:
      DEBUG_LOG("Fetching revision properties for r%d" % (revision))
      def _authz_callback(fsroot, path, pool, rootname=rootname):
        # Authz check - we always pass True for 'is_file' so we can
        # avoid recursive readability checks.
        return self._is_readable(rootname, path,
                                 svn.fs.revision_root_revision(fsroot),
                                 True) and 1 or 0
      proplist = svn.repos.fs_revision_proplist(repo, revision, _authz_callback)
    DEBUG_LOG("Completed fetching properties (found %d)" % (len(proplist)))
    return proplist

  def _get_symlink_target(self, fsroot, path):
    props = svn.fs.node_proplist(fsroot, path)
    if not props.has_key(svn.core.SVN_PROP_SPECIAL):
      return None
    pathspec = ''
    ### FIXME: We're being a touch sloppy here, only checking the first line
    ### of the file.
    stream = svn.fs.file_contents(fsroot, path)
    try:
      pathspec, eof = svn.core.svn_stream_readline(stream, '\n')
    finally:
      svn.core.svn_stream_close(stream)
    if pathspec[:5] != 'link ':
      return None
    return pathspec[5:]

  def _resolve_local_symlink_target(self, link_path, target_path):
    # Absolute targets don't get resolved.
    if target_path[0] == '/':
      return None

    # Split the link and target into path components, removing empty
    # and '.'  components.
    link_pieces = filter(lambda x: (len(x) > 0 and x != '.'),
                         link_path.split('/'))
    target_pieces = filter(lambda x: (len(x) > 0 and x != '.'),
                           target_path.split('/'))

    # Empty links and targets don't get resolved.
    if not (len(link_pieces) and len(target_pieces)):
      return None

    # This is a relative target, so join it with the link dirname.  We
    # insert an updir (..) between the segments to effectively use the
    # dirname of the link itself as the base path; our processing
    # below will work all that out.
    pieces = link_pieces + ['..'] + target_pieces

    # Allow updirs to swallow the preceding components (if any;
    # otherwise, bail out).
    i = 0
    while i < len(pieces):
      if pieces[i] == '..':
        if i:
          del pieces[i-1:i+1]
          i = i - 1
        else:
          return None
      else:
        i = i + 1

    # Return the reassembled pieces.
    return '/'.join(pieces)

  def _path_at_revision(self, fs, rootname, peg_revision, path,
                        node_kind, revision):
    # Find PATH@PEG_REVISION as it exists in REVISION.  Ideally,
    # we'd use a real authz callback function like this:
    #
    #   def _authz_callback(fsroot, authz_path, pool, rootname=rootname):
    #     return self._is_readable(rootname, authz_path,
    #                              svn.fs.revision_root_revision(fsroot),
    #                              node_kind) and 1 or 0
    #
    # Unfortunately, while composing this code cmpilato discovered a
    # security flaw in Subversion's svn_repos_trace_node_locations()
    # implementation.  This means we'll need to do our own authz
    # checks on the output from this method.
    node_locs = svn.repos.trace_node_locations(fs, path, peg_revision,
                                               [revision], None)
    if revision in node_locs:
      # Authz check.
      if not self._is_readable(rootname, node_locs[revision], revision,
                               node_kind == svn.core.svn_node_file):
        raise SvnAPIPathNotFoundError()
      return node_locs[revision]
    else:
      raise SvnAPIPathNotFoundError()

  # ----------- #
  # JS Objects  #
  # ----------- #

  def _make_revmeta(self, revision, revprops):
    revprops = self._filter_props(revprops)
    author = _pop(revprops, svn.core.SVN_PROP_REVISION_AUTHOR)
    log    = _pop(revprops, svn.core.SVN_PROP_REVISION_LOG)
    date   = _pop(revprops, svn.core.SVN_PROP_REVISION_DATE)
    if date is not None:
      secs = svn.core.svn_time_from_cstring(date) / 1000000;
      date = time.strftime('%Y-%m-%dT%H:%M:%SZ', time.gmtime(secs))
    return  {
      'revision': revision,
      'author': author,
      'description': log,
      'whenDate': date,
      'otherProperties': revprops,
      }

  # ----------- #
  # API Methods #
  # ----------- #

  def roots(self, params):
    response = {
      'roots': None,
      'error': None,
      }
    try:
      ##
      ## THIS METHOD DOES *NOT* PERFORM PATH-BASED AUTHORIZATION
      ## CHECKS.  DO NOT EXPOSE THROUGH THE PUBLIC API!
      ##
      roots = []
      for item in os.listdir(self.svn_parent_path):
        fullpath = os.path.join(self.svn_parent_path, item)
        if (os.path.isdir(fullpath) and
            os.path.isdir(os.path.join(fullpath, 'db')) and
            os.path.isfile(os.path.join(fullpath, 'format')) and
            os.path.isfile(os.path.join(fullpath, 'db', 'fs-type'))):
          try:
            repo = self._getrepo(item)
            fs = svn.repos.fs(repo)
            youngest = self._getrev(fs, -1)
            revprops = svn.fs.revision_proplist(fs, youngest)
            roots.append({
              'name': item,
              'revMeta': self._make_revmeta(youngest, revprops),
              })
          except:
            pass # Not a Subversion repository
      response.update({
        'roots': roots,
        })
    except Exception, e:
      response.update({
        'error': _json_error(e),
        })
    finally:
      return response

  def listdir(self, params):
    response = {
      'error': None,
      'entries': None,
      'revMetaKey': None,
      'revMetas': {},
      'revPath' : None,
      }
    try:
      # Parse parameters.
      rootname = params['repo_name']
      repo = self._getrepo(rootname)
      fs = svn.repos.fs(repo)
      revision = self._getrev(fs, params['revision'])
      fsroot = self._getroot(fs, revision)
      path = self._getpath(params['path'])
      collapse_empty_dirs = False
      if self.allow_collapse_empty_dirs:
        collapse_empty_dirs = params.get('collapse_empty_dirs')

      DEBUG_LOG("Beginning listdir('%s', r%d)" % (path, revision))

      # Authz check
      if not self._is_readable(rootname, path, revision, False):
        raise SvnAPIPathNotFoundError()

      error = None
      dirents = []
      created_rev = svn.fs.node_created_rev(fsroot, path)
      # The path may differ at the created_rev due to renames.
      created_path = svn.fs.node_created_path(fsroot, path)[1:]
      revmeta_key = str(created_rev)
      revprops = self._get_revprops(rootname, repo, created_rev)
      revmetas = {
        revmeta_key: self._make_revmeta(created_rev, revprops),
        }

      deadline = time.time() + self.listdir_timeout
      for child, dirent in svn.fs.dir_entries(fsroot, path).items():
        this_path = path + '/' + child

        # If we've taken too long to harvest information, we'll just
        # roll with what we've got and flag the fact that we hit a
        # timeout.
        if time.time() > deadline:
          error = _json_error(SvnAPIAuthorizationTimeoutError())
          break

        # Authz check - skip/omit unreadable children
        if not self._is_readable(rootname, this_path, revision,
                                 dirent.kind == svn.core.svn_node_file):
          continue

        if dirent.kind == svn.core.svn_node_dir and collapse_empty_dirs:
          # see if this directory needs to be crawled for (deep) children.
          json_dirent = self._bypass_empty_tree(rootname, fsroot, path, child)
        else:
          # add this entry to our return list.
          this_crpath = svn.fs.node_created_path(fsroot, this_path)[1:]
          this_rev = svn.fs.node_created_rev(fsroot, this_path)
          json_dirent = {
            'name': child,
            'parentTree': None,
            'kind': self._node_kind_string(dirent.kind),
            'revMetaKey': str(this_rev),
            'revPath': this_crpath,
            }
        if dirent.kind == svn.core.svn_node_file:
          symlink_target = self._get_symlink_target(fsroot, this_path)
          if symlink_target is not None:
            # See if the symlink target is a readable path within this
            # same repository.  If so, we'll return its repository
            # path and kind.
            target_repos_path = self._resolve_local_symlink_target(this_path,
                                                                   symlink_target)
            if target_repos_path is not None:
              target_repos_kind = svn.fs.check_path(fsroot, target_repos_path)
              is_file = (target_repos_kind == svn.core.svn_node_file)
              if ((target_repos_kind == svn.core.svn_node_none) or
                  (not self._is_readable(rootname, target_repos_path,
                                         revision, is_file))):
                target_repos_path = None
            if target_repos_path is not None:
              json_dirent['symlinkInfo'] = {
                'targetPath': symlink_target,
                'targetReposPath': target_repos_path,
                'targetKind': self._node_kind_string(target_repos_kind),
                }
            else:
              json_dirent['symlinkInfo'] = {
                'targetPath': symlink_target,
                }
        this_revmeta_key = json_dirent['revMetaKey']
        if not revmetas.has_key(this_revmeta_key):
          this_rev = int(this_revmeta_key)
          revprops = self._get_revprops(rootname, repo, this_rev)
          revmetas[this_revmeta_key] = self._make_revmeta(this_rev, revprops)
        dirents.append(json_dirent)

      def _relpath(dirent):
        return os.path.join(dirent.get('parentTree') or '',
                            dirent.get('name'))
      dirents.sort(lambda a, b: _compare_paths(_relpath(a), _relpath(b)))
      response.update({
        'entries': dirents,
        'revMetaKey': revmeta_key,
        'revMetas': revmetas,
        'revPath' : created_path,
        'error': error,
        })
    except Exception, e:
      response.update({
        'error': _json_error(e),
        })
    finally:
      DEBUG_LOG("Finished listdir()")
      return response

  def getfileraw(self, params):
    try:
      # Parse parameters.
      rootname = params['repo_name']
      repo = self._getrepo(rootname)
      fs = svn.repos.fs(repo)
      revision = self._getrev(fs, params['revision'])
      fsroot = self._getroot(fs, revision)
      path = self._getpath(params['path'])
      skip_encoding_detection = params['skip_encoding_detection']

      # Authz check
      if not self._is_readable(rootname, path, revision, True):
        raise SvnAPIPathNotFoundError()

      # Get content, size, MIME info, etc.
      scp = SvnContentProxy(fsroot, path, max_size=0)
      mime_type, encoding = _get_mime_data(fsroot, path, True)
      file_size = svn.fs.file_length(fsroot, path)
      content = None

      # Unless we're told not to bother, try to determine the file's
      # encoding, either from the Subversion properties or, for files
      # smaller than our memory threshold, by inspection.  If we can
      # determine an encoding, we'll claim the file is plaintext in
      # that encoding, regardless of what the detected/versioned MIME
      # type claims.
      if ((not encoding)
          and (not skip_encoding_detection)
          and (file_size <= MAX_FILE_LENGTH)):
        content = scp.get_content()
        import SourceForge
        encoding = SourceForge.detect_encoding(content)
      if encoding:
        mime_type = 'text/plain; charset=%s' % (encoding)

      # Ideally, we'd use only either the precalculated 'content'
      # string or 'scp.get_generator()', but BerkeleyDB-backed
      # repositories seem to struggle with the generator logic (to the
      # point of SEFAULTing the server child process).  See
      # artf267432.
      if content is not None:
        fileContent = content
      elif svn.fs.type(svn.fs.path(fs)) == 'bdb':
        if file_size > MAX_FILE_LENGTH:
          raise SvnAPIFileTooLargeError()
        else:
          fileContent = scp.get_content()
      else:
        fileContent = scp.get_generator()

      return {
        'fileContent': fileContent,
        'fileSize': file_size,
        'mimeType': mime_type,
        }
    except svn.core.SubversionException, e:
      _svn2api_error(e)
    except Exception, e:
      raise

  def getfile(self, params):
    response = {
      'revMeta': None,
      'fileContent': None,
      'mimeType': None,
      'error': None,
      }
    try:
      # Parse parameters.
      rootname = params['repo_name']
      repo = self._getrepo(rootname)
      fs = svn.repos.fs(repo)
      revision = self._getrev(fs, params['revision'])
      fsroot = self._getroot(fs, revision)
      path = self._getpath(params['path'])

      # Authz check
      if not self._is_readable(rootname, path, revision, True):
        raise SvnAPIPathNotFoundError()

      created_rev = svn.fs.node_created_rev(fsroot, path)
      # The path may differ at the created_rev due to renames.
      created_path = svn.fs.node_created_path(fsroot, path)[1:]
      revprops = self._get_revprops(rootname, repo, created_rev)
      revmeta = self._make_revmeta(created_rev, revprops)
      response.update({
        'revMeta': revmeta,
        'revPath': created_path,
        })

      if svn.fs.file_length(fsroot, path) > MAX_FILE_LENGTH:
        raise SvnAPIFileTooLargeError()

      contents = SvnContentProxy(fsroot, path).get_content()
      mime_type, encoding = _get_mime_data(fsroot, path, True)
      response.update({
        'mimeType': mime_type,
        })
      import SourceForge
      if not encoding:
        encoding = SourceForge.detect_encoding(contents)
      try:
        contents = SourceForge.transcode_to_utf8(contents, encoding)
      except:
        raise SvnAPIUnsupportedContentEncodingError()
      response.update({
        'fileContent': contents,
        })
    except Exception, e:
      response.update({
        'error': _json_error(e),
        })
    finally:
      return response

  def getfilelines(self, params):
    response = {
      'revMetas': {},
      'revMetaKey': None,
      'fileContentLines': [],
      'mimeType': None,
      'truncationWarning': None,
      'error': None,
      }
    try:
      # Parse parameters.
      rootname = params['repo_name']
      repo = self._getrepo(rootname)
      fs = svn.repos.fs(repo)
      revision = self._getrev(fs, params['revision'])
      fsroot = self._getroot(fs, revision)
      path = self._getpath(params['path'])
      start_line = params.get('start_line', 0)
      end_line = params.get('end_line', -1)
      allow_truncate = params.get('allow_truncate')

      # Authz check
      if not self._is_readable(rootname, path, revision, True):
        raise SvnAPIPathNotFoundError()

      created_rev = svn.fs.node_created_rev(fsroot, path)
      # The path may differ at the created_rev due to renames.
      created_path = svn.fs.node_created_path(fsroot, path)[1:]
      revmeta_key = str(created_rev)
      revprops = self._get_revprops(rootname, repo, created_rev)

      ### FIXME: Dear future me, we really don't need a dictionary of
      ### revMetas with a key into it.  We just need a single revMeta
      ### here.  There is precedent.  Love, old-and-forgetful me.
      revmetas = {
        revmeta_key: self._make_revmeta(created_rev, revprops)
        }
      response.update({
        'revMetas': revmetas,
        'revMetaKey': revmeta_key,
        'revPath': created_path,
        })

      # If we're not allowing truncation and the file is too large,
      # raise an error.
      if (not allow_truncate and
          svn.fs.file_length(fsroot, path) > MAX_FILE_LENGTH):
        raise SvnAPIFileTooLargeError()

      scp = SvnContentProxy(fsroot, path)
      lines = scp.get_lines(start_line, end_line)

      # If we've peeked into more lines of the file than are allowed,
      # raise an error.
      if (start_line + len(lines)) > MAX_FILE_LINES:
        raise SvnAPIFileTooManyLinesError()

      truncated = scp.was_truncated()
      mime_type, encoding = _get_mime_data(fsroot, path, True)
      response.update({
        'mimeType': mime_type,
        })

      import SourceForge
      if lines and not encoding:
        encoding = SourceForge.detect_encoding(lines[0])
      try:
        for i in range(len(lines)):
          lines[i] = {
            'text': SourceForge.transcode_to_utf8(lines[i], encoding),
            'lineNum': start_line + i + 1,
            }
      except:
        raise SvnAPIUnsupportedContentEncodingError()

      response.update({
        'fileContentLines': lines,
        })

      # If we were unable to fetch all the lines the caller wanted,
      # then add a warning.
      if truncated and ((end_line - start_line) > len(lines)):
        response.update({
          'truncationWarning': _json_error(SvnAPIFileTruncatedError()),
          })

    except Exception, e:
      response.update({
        'error': _json_error(e),
        })
    finally:
      return response

  def blame(self, params):
    response = {
      'revMetas': {},
      'blameLines': [],
      'error': None,
      }
    try:
      # Parse parameters.
      rootname = params['repo_name']
      repo = self._getrepo(rootname)
      fs = svn.repos.fs(repo)
      revision = self._getrev(fs, params['revision'])
      fsroot = self._getroot(fs, revision)
      path = self._getpath(params['path'])

      # Authz check
      if not self._is_readable(rootname, path, revision, True):
        raise SvnAPIPathNotFoundError()

      revmetas = {}
      response.update({
        'revMetas': revmetas,
        })

      annotator = SvnAnnotator(self._geturl(rootname, path), revision)

      # Update the revMeta map with any revisions found by the
      # annotator.
      def _revision_to_revmeta_map(rev):
        revprops = self._get_revprops(rootname, repo, rev)
        return {
          str(rev): self._make_revmeta(rev, revprops),
          }
      map(lambda x: revmetas.update(_revision_to_revmeta_map(x)),
          annotator.revisions())

      response.update({
        'blameLines': annotator.json_lines(),
        })

    except Exception, e:
      response.update({
        'error': _json_error(e),
        })
    finally:
      return response

  def getprops(self, params):
    response = {
      'props': {},
      'propsDropped': False,
      'error': None,
      }
    try:
      # Parse parameters.
      rootname = params['repo_name']
      repo = self._getrepo(rootname)
      fs = svn.repos.fs(repo)
      revision = self._getrev(fs, params['revision'])
      params_path = params.get('path')

      DEBUG_LOG("Beginning getprops('%s', r%d)"
                % (params_path or "(None)", revision))

      if params_path is not None:
        # Node props
        path = self._getpath(params_path)
        fsroot = self._getroot(fs, revision)
        node_kind = svn.fs.check_path(fsroot, path)

        # Authz check
        if not self._is_readable(rootname, path, revision,
                                 node_kind == svn.core.svn_node_file):
          raise SvnAPIPathNotFoundError()

        props = svn.fs.node_proplist(fsroot, path)
      else:
        # Revision props
        props = svn.fs.revision_proplist(fs, revision)

      filtered = self._filter_props(props)
      response.update({
        'props': filtered,
        'propsDropped': len(filtered) < len(props),
        })
    except Exception, e:
      response.update({
        'error': _json_error(e),
        })
    finally:
      DEBUG_LOG("Finished getprops()")
      return response

  def gethistory(self, params):
    response = {
      'history': [],
      'kind': None,
      'error': None,
      }
    try:
      # Parse parameters.
      rootname = params['repo_name']
      repo = self._getrepo(rootname)
      fs = svn.repos.fs(repo)
      revision = self._getrev(fs, params['revision'])
      fsroot = self._getroot(fs, revision)
      path = self._getpath(params['path'])
      numlogs = params['numlogs']

      node_kind = svn.fs.check_path(fsroot, path)

      # Authz check
      if not self._is_readable(rootname, path, revision,
                               node_kind == svn.core.svn_node_file):
        raise SvnAPIPathNotFoundError()

      receiver = SvnHistoryReceiver(fs, numlogs)
      histories = []
      svn.repos.history2(fs, path, receiver.add_history, None, revision,
                         len(path) and 1 or 0, 1)
      for hist_rev, hist_path in receiver:

        # Authz check - if we walk back to an unreadable location, we
        # stop walking.
        if not self._is_readable(rootname, hist_path, hist_rev,
                                 node_kind == svn.core.svn_node_file):
          break

        revprops = self._get_revprops(rootname, repo, hist_rev)
        history = {
          'reposPath': hist_path,
          'revMeta': self._make_revmeta(hist_rev, revprops),
          }
        histories.append(history)

      response.update({
        'history': histories,
        'kind': self._node_kind_string(node_kind),
        })
    except Exception, e:
      response.update({
        'error': _json_error(e),
        })
    finally:
      return response

  def lastdiff(self, params):
    response = {
      'diff': None,
      'kind': None,
      'error': None,
      }
    try:
      # Parse parameters.
      rootname = params['repo_name']
      repo = self._getrepo(rootname)
      fs = svn.repos.fs(repo)
      revision = self._getrev(fs, params['revision'])
      fsroot = self._getroot(fs, revision)
      path = self._getpath(params['path'])
      unified = params['type'] == 'unified'
      context_lines = params.get('context_lines', 3)
      base_revision = params.get('base_revision', -1)

      node_kind = svn.fs.check_path(fsroot, path)

      # Authz check
      if not self._is_readable(rootname, path, revision,
                               node_kind == svn.core.svn_node_file):
        raise SvnAPIPathNotFoundError()

      if base_revision != -1:
        if base_revision < 0 or base_revision > revision:
          raise SvnAPIUnsupportedOperation("Invalid base revision")
        try:
          prev_hist_path = self._path_at_revision(fs, rootname, revision, path,
                                                  node_kind, base_revision)
          prev_fsroot = self._getroot(fs, base_revision)
        except Exception, e:
          prev_fsroot = prev_hist_path = None
      else:
        receiver = SvnHistoryReceiver(fs, 2)
        svn.repos.history2(fs, path, receiver.add_history, None, revision, 1, 1)
        last_hist_rev, last_hist_path = receiver[0]
        if last_hist_rev != revision:
          raise SvnAPIPathNotChangedError()
        try:
          prev_hist_rev, prev_hist_path = receiver[1]

          # Authz check - ensure that the path we're comparing against
          # is also readable.  If not, this diff needs to look like a
          # fresh addition (diff against the empty data set).
          assert self._is_readable(rootname, prev_hist_path, prev_hist_rev,
                                   node_kind == svn.core.svn_node_file)
          prev_fsroot = self._getroot(fs, prev_hist_rev)
        except:
          prev_fsroot = prev_hist_path = None

      if (prev_fsroot and
          ((node_kind == svn.core.svn_node_file) and
           (svn.fs.file_length(fsroot, path) > MAX_FILE_LENGTH or
            svn.fs.file_length(prev_fsroot, prev_hist_path) > MAX_FILE_LENGTH))):
        raise SvnAPIFileTooLargeError()

      differ = SvnDifferencer(fsroot, path, prev_fsroot, prev_hist_path,
                              node_kind, unified, context_lines)
      response.update({
        'kind': self._node_kind_string(node_kind),
        'diff': differ.get_delta(),
        })
    except Exception, e:
      response.update({
        'error': _json_error(e),
        })
    finally:
      return response

  def revision(self, params):
    response = {
      'revMeta': None,
      'changedPaths': None,
      'error': None,
      }
    try:
      # Parse parameters.
      rootname = params['repo_name']
      repo = self._getrepo(rootname)
      fs = svn.repos.fs(repo)
      revision = self._getrev(fs, params['revision'])
      fsroot = self._getroot(fs, revision)

      revprops = self._get_revprops(rootname, repo, revision)
      cc = svn.repos.ChangeCollector(fs, fsroot)
      e_ptr, e_baton = svn.delta.make_editor(cc)
      svn.repos.svn_repos_replay(fsroot, e_ptr, e_baton)
      changes = []
      for changed_path, change in cc.get_changes().items():
        # Authz check - omit unreadable paths.  We always pass True
        # for 'is_file' so we can avoid recursive readability checks.
        if not self._is_readable(rootname, changed_path, revision, True):
          continue

        action = ({ svn.repos.CHANGE_ACTION_MODIFY:  'modify',
                    svn.repos.CHANGE_ACTION_ADD:     'add',
                    svn.repos.CHANGE_ACTION_DELETE:  'delete',
                    svn.repos.CHANGE_ACTION_REPLACE: 'replace',
                    }).get(change.action, 'unknown')
        copyfrom_path = copyfrom_rev = None
        if action in ('add', 'replace'):
          copyfrom_rev, copyfrom_path = svn.fs.copied_from(fsroot, change.path)

        # Authz check - if a copyfrom location isn't readable, claim
        # the operation isn't a copy.  Here again, we pass True for
        # 'is_file' to avoid recursive checks.
        if copyfrom_path and copyfrom_rev >= 0:
          if not self._is_readable(rootname, copyfrom_path,
                                   copyfrom_rev, True):
            copyfrom_path = None
            copyfrom_rev = -1

        changes.append({
          'reposPath': change.path,
          'kind': self._node_kind_string(change.item_kind),
          'action': action,
          'copySourcePath': copyfrom_path,
          'copySourceRev': copyfrom_rev,
          })
      response.update({
        'revMeta': self._make_revmeta(revision, revprops),
        'changedPaths': changes,
        })
    except Exception, e:
      response.update({
        'error': _json_error(e),
        })
    finally:
      return response

  def pathatrev(self, params):
    response = {
      'error': None,
      }
    try:
      # Parse parameters.
      rootname = params['repo_name']
      repo = self._getrepo(rootname)
      fs = svn.repos.fs(repo)
      revision = self._getrev(fs, params['revision'])
      peg_revision = self._getrev(fs, params['peg_revision'])
      params_path = params.get('path')
      path = self._getpath(params_path)

      node_kind = svn.fs.check_path(self._getroot(fs, peg_revision), path)
      if not self._is_readable(rootname, path, peg_revision,
                               node_kind == svn.core.svn_node_file):
        raise SvnAPIPathNotFoundError()

      revpath = self._path_at_revision(fs, rootname, peg_revision, path,
                                       node_kind, revision)
      response.update({
        'path': revpath,
        'created_rev': revision,
      })
    except Exception, e:
      response.update({
        'error': _json_error(e),
        })
    finally:
      return response


class SvnAPIError(Exception):
  """Subversion API custom error."""
  def __init__(self, code, message):
    self.code = code
    self.message = message

  def to_json(self):
    return _make_json_error(self.code, self.message, _stacktrace())


class SvnAPIGenericError(SvnAPIError):
  """Uh-oh"""
  def __init__(self, message):
    SvnAPIError.__init__(self, 'UNKNOWN_ERROR', message)


class SvnAPIPathNotFoundError(SvnAPIError):
  """Path not found"""
  def __init__(self):
    SvnAPIError.__init__(self,
                         'PATH_NOT_FOUND',
                         'Path not found')


class SvnAPIRevisionNotFoundError(SvnAPIError):
  """Path not found"""
  def __init__(self):
    SvnAPIError.__init__(self,
                         'REVISION_NOT_FOUND',
                         'Revision not found')


class SvnAPIPathNotChangedError(SvnAPIError):
  """Path not modified in the specified revision"""
  def __init__(self):
    SvnAPIError.__init__(self,
                         'PATH_NOT_CHANGED',
                         'Path not modified in the specified revision')


class SvnAPIUnsupportedContentEncodingError(SvnAPIError):
  """Unsupported file content encoding"""
  def __init__(self):
    SvnAPIError.__init__(self,
                         'ENCODING_ERROR',
                         'Unsupported file content encoding')


class SvnAPIFileTooLargeError(SvnAPIError):
  """File larger than configured maximum size"""
  def __init__(self):
    SvnAPIError.__init__(self,
                         'FILE_TOO_LARGE',
                         'File larger than configured maximum ('
                         + str(MAX_FILE_LENGTH_MB) +
                         'MB) for this operation')


class SvnAPIFileTooManyLinesError(SvnAPIError):
  """File contains more lines than configured maximum"""
  def __init__(self):
    SvnAPIError.__init__(self,
                         'FILE_TOO_MANY_LINES',
                         'File contains more lines than configured '
                         + 'maximum (' + str(MAX_FILE_LINES) +
                         ') for this operation')


class SvnAPIFileTruncatedError(SvnAPIError):
  """File larger than configured maximum size; contents truncated"""
  def __init__(self):
    SvnAPIError.__init__(self,
                         'FILE_TOO_LARGE_TRUNCATED',
                         'File larger than configured maximum ('
                         + str(MAX_FILE_LENGTH_MB) +
                         'MB) for this operation; contents truncated')


class SvnAPIAnnotationAlignmentError(SvnAPIError):
  """Unable to align file content with annotation data"""
  def __init__(self, reason):
    SvnAPIError.__init__(self,
                         'ANNOTATION_ALIGNMENT_ERROR',
                         'Unable to annotate file content: %s' % (reason))


class SvnAPIAnnotationBinaryFileError(SvnAPIError):
  """Unable to align file content with annotation data"""
  def __init__(self):
    SvnAPIError.__init__(self,
                         'ANNOTATION_BINARY_FILE_ERROR',
                         'Unable to annotate file content: binary file content')

class SvnAPIAuthorizationTimeoutError(SvnAPIError):
  """Timed out authorizing access to one or more paths"""
  def __init__(self):
    SvnAPIError.__init__(self,
                         'AUTHORIZATION_TIMEOUT',
                         'Timed out authorizing access to one or more paths')

class SvnAPIUnsupportedOperation(SvnAPIError):
  """The requested operation is unsupported"""
  def __init__(self, details=None):
    SvnAPIError.__init__(self,
                         'UNSUPPORTED_OPERATION',
                         'The requested operation is unsupported%s'
                         % (details and ': ' + details or ''))
