from app.scripts import BaseMigrationScript
import os
import app.util

class MigrationScript(BaseMigrationScript):

    scriptDescription = "Create symbolic link of CVS repository base location"

    def migrate(self):

        scmRepoDataDir = self.cfg['SCM_REPOSITORY_DATA_DIR']

        if not os.path.isdir(scmRepoDataDir):
            cmd = "mkdir -p %s" % scmRepoDataDir
            app.util.executeCommand(cmd)

        self.createLinkToCVSPath(scmRepoDataDir)

    def createLinkToCVSPath(self, scmDataDir):

        # Softlink will be created to an new SCM data directory while
        # CVS_REPOSITORY_BASE token is referring to an old location like
        # /cvsroot

        cvsRepoRootPath = self.cfg['CVS_REPOSITORY_BASE']
        cvsSymlinkPath = "%s/cvsroot" % scmDataDir

        # In upgrade scenario The installer framework will create folder under /opt/collabnet/teamforge/var/scm/cvsroot
        # before calling migration hooks but which needs to be removed and created as soft link.
        if cvsSymlinkPath != cvsRepoRootPath and not os.path.islink(cvsSymlinkPath) and \
                os.path.isdir(cvsSymlinkPath) and not os.listdir(cvsSymlinkPath):
                os.rmdir(cvsSymlinkPath)
        if os.path.exists(cvsSymlinkPath):
            return
        self.createSoftLink(cvsRepoRootPath, cvsSymlinkPath)

    def createSoftLink(self, originalPath, symlinkPath):

        scmUser = self.cfg["HTTPD_USER"]
        scmGroup = self.cfg["HTTPD_GROUP"]
        cmd = "chown -h %s:%s %s"  % (scmUser, scmGroup, symlinkPath)

        os.symlink(originalPath, symlinkPath)
        app.util.runSystem(cmd, self.log)

