#!/usr/bin/env python2
#
#
#
#
# The above four blank lines brought to you courtesy of the
# CEE wrapper trickery.
# -*- Python -*-

import os
import re
import sys
import pwd
import threading
import time
import getopt
import select
import app.util
from app import adapter_svn

THREADS = 5
lock = threading.Lock()

num_checked = 0
num_skipped = 0
num_healthy = 0
problematic_repos = []

#dictionary with key being thread index(0, 1, 2...etc) value being list of
#repositories
repo_per_thread = {}

cfg = app.util.getRuntimeConfiguration()

class CheckRepositoryThread(threading.Thread):
    def updateGlobalStats(self):
        lock.acquire()
        try:
            global num_skipped, num_checked, num_healthy, problematic_repos
            num_skipped = num_skipped + self.num_skipped
            num_checked = num_checked + self.num_checked
            num_healthy = num_healthy + self.num_healthy
            for reposName in self.problematic_repos:
                problematic_repos.append(reposName)
        finally:
            lock.release()

    # Executes the command and waits for output/errput untill TIMEOUT.
    # If no error is found returns 0, else returns 1.
    # If no response within TIMEOUT returns 1.
    # When it returns 1 it writes the reason to STDERR.
    def execWithTimeout(self, command, timeout=10):
        infile, outfile, errfile = os.popen3(command)
        infile.close()
        outfd = outfile.fileno()
        errfd = errfile.fileno()
        ready = select.select([outfd, errfd], [], [], timeout)
        if errfd in ready[0]:
            err_lines = errfile.readlines()
            sys.stderr.write("Error running '%s':\n   %s\n" \
                              % (command, '   '.join(err_lines)))
            return 1
        elif outfd in ready[0]:
            #Should we read this output, may be to free of the buffers?
            out_lines = outfile.readlines()
            return 0
        else:
            sys.stderr.write("Timeout(%d) running '%s'\n"
                             % (timeout, command))
            return 1

    def run(self):
        self.num_checked = 0
        self.num_skipped = 0
        self.num_healthy = 0
        self.problematic_repos = []
        global repo_per_thread
        svn_binary = os.path.join(cfg.get('SVN_BINARY'))
        try:
            repos_to_check = repo_per_thread[self.getName()]
        except KeyError:
            return
        for reposName in repos_to_check:
            self.num_checked = self.num_checked + 1
            if not os.path.isdir(os.path.join(reposName, 'db')):
                sys.stderr.write(
                    "'%s' is not a Subversion repository.  "
                    "Skipping sanity check.\n" % (reposName))
                self.num_skipped = self.num_skipped + 1
                continue

            command = '%s ls "file://%s"' % (svn_binary, reposName)
            retVal = self.execWithTimeout(command)
            if retVal:
                self.problematic_repos.append(reposName)
            else:
                self.num_healthy = self.num_healthy + 1

        del repo_per_thread[self.getName()]
        self.updateGlobalStats()

def printSummary():
    print 'Summary:'
    print 'Number of repositories checked: ', num_checked
    print 'Number of repositories skipped: ', num_skipped
    print 'Number of healthy repositories: ', num_healthy
    if len(problematic_repos):
        print 'List of broken repositories:'
        for broken_repo in problematic_repos:
            print broken_repo

def usageAndExit(exitValue):
    sys.stderr.write("""Usage: %s [--help] [--threads=%d] [--repositories=/path/to/file/having/repositories/in/each/line]

Unless --help is provided (which shows this output and does nothing else),
check the sanity of all the repositories in the live repository area of
the site.

WARNING:  This program access repositories directly, and therefore must be
run from the Subversion service host machine!!

It checks %d repositories at a time.
""" % (os.path.basename(sys.argv[0]), THREADS, THREADS))
    sys.exit(exitValue)

def main():
    try:
        adapter_svn.checkSafeUser(cfg)
    except Exception, e:
        print >> sys.stderr, str(e)
        sys.exit(1)

    os.environ['HOME'] = pwd.getpwuid(os.getuid())[5]
    threads = THREADS
    # path to file having repository names in each line.
    repository_list_file = None
    try:
        opts, args = getopt.getopt(sys.argv[1:], 'h?t=r=',
                                   ['help',
                                    'threads=',
                                    'repositories='
                                   ])
    except getopt.GetoptError, e:
        print e
        usageAndExit(1)
    for opt, value in opts:
        if opt == '-h' or opt == '--help' or opt == '-?':
            usageAndExit(0)
        elif opt == '-t' or opt == '--threads':
            threads = int(value)
        elif opt == '-r' or opt == '--repositories':
            repository_list_file = value

    if repository_list_file:
        f = file(repository_list_file)
        repositories = f.read().split('\n')
        clean_repositories = []
        for repo in repositories:
            if repo:
                clean_repositories.append(repo)
        repositories = clean_repositories
    else:
        repositories = []
        for repos_home in ("SUBVERSION_REPOSITORY_BASE",
                           "SUBVERSION_BRANDING_REPOSITORY_BASE",
                           ):
            repo_home = cfg.get(repos_home)
            for repo in os.listdir(repo_home):
                if os.path.isdir(os.path.join(repo_home, repo)):
                    repo_path = os.path.join(repo_home, repo)
                    repositories.append(repo_path)

    total_repositories = len(repositories)
    num_repo_per_thread = (total_repositories/threads) + 1
    repo_counter = 0
    thread_counter = 0
    while repo_counter < total_repositories:
        repo_per_thread[str(thread_counter)] = \
              repositories[repo_counter:repo_counter+num_repo_per_thread]
        repo_counter = repo_counter + num_repo_per_thread
        thread_counter = thread_counter + 1
    for x in range(threads):
        worker = CheckRepositoryThread()
        worker.setName(x)
        worker.start()
    while len(repo_per_thread) > 0:
        time.sleep(1)
    printSummary()

if __name__ == "__main__":
    main()
