#!/bin/bash

# Parse through the command line options
while getopts hd: opt
  do 
    case $opt in 
     # Get the dir containing repo(s), as an abspath
     d) REPO_DIR=`echo $OPTARG | sed -e 's/\/*$//'`;;
     h|[?]) echo >&2 "Usage: `basename $0` [-d repo|directory containing the repo(s)]
                           [-h] Prints this help message"
          exit 1;;
    esac
done
shift `expr $OPTIND - 1`

RUNTIME_DIR=`grep app.runtime /etc/sourceforge.properties | cut -f2 -d=`
# Obtain the svn wrapper binaries, if available
SVNADMIN=`eval "\`$RUNTIME_DIR/scripts/options-to-shell-variables.py\`"; echo $SVNADMIN_BINARY`
SVNLOOK=`eval "\`$RUNTIME_DIR/scripts/options-to-shell-variables.py\`"; echo $SVNLOOK_BINARY` 

# The method which calls the bdb-fsfs.sh to perform the logic
convert_to_fsfs ()
{
  REPO=$1
  if [ `cat $REPO/db/fs-type` == 'bdb' ]; 
    then
     # Check whether the repo has three dir structure and skip, if so
      grep set_data_dir $REPO/db/DB_CONFIG > /dev/null 2>&1
      if [ $? -ne 0 ];
      then
        SVNADMIN_BINARY=$SVNADMIN SVNLOOK_BINARY=$SVNLOOK $RUNTIME_DIR/scripts/bdb-fsfs.sh $REPO
      fi
      else
        echo -e "`basename $REPO` is already of type fsfs\n"
  fi
}

# Invoke the base script, which performs the logic
if [ -d $REPO_DIR ];
then
  echo "Start -- convert to fsfs"

  if [ -e "$REPO_DIR/db/fs-type" ]; 
  then 
    REPO_ABS_PATH="$REPO_DIR"
    # Convert the repo directly passed as argument to fsfs
    convert_to_fsfs $REPO_ABS_PATH
  else
    # Iterate over all the bdb repos and convert them to fsfs
    for REPO in `ls $REPO_DIR`; do
      if [ -e "$REPO_DIR/$REPO/db/fs-type" ];
      then
        REPO_ABS_PATH="$REPO_DIR/$REPO"
      else 
        echo "Invalid directory argument"
        exit 1;
      fi
      
      # Convert every repo under $REPO_DIR to fsfs
      convert_to_fsfs $REPO_ABS_PATH
    done
  fi
fi 
