#!/bin/bash

REPO=$1
SVNADMIN=${SVNADMIN_BINARY-`which svnadmin`}
SVNLOOK=${SVNLOOK_BINARY-`which svnlook`}
REPO_LOCATION=`dirname $REPO`
HTTPD_USER=httpd
HTTPD_GROUP=$HTTPD_USER

# Convert bdb repo to fsfs using dump&load mechanism:
 # Move the repo as a temp one so as to take a dump of it
TEMP_REPO=$REPO.`date '+%F%H:%M:%S'`.$$
REPO_NAME=`basename $REPO`

REV_BEFORE_DUMP=`su - $HTTPD_USER -c "$SVNLOOK youngest $REPO"`
if [ ! -f /tmp/$REPO_NAME.dump ];
then 
  su - $HTTPD_USER -c "mv $REPO $TEMP_REPO"
  if [ $? -ne 0 ]; 
  then
    echo -e "Renaming the repo failed, check for the proper permissions\n"
    exit $?
  fi

  su - $HTTPD_USER -c "$SVNADMIN dump -q $TEMP_REPO > /tmp/$REPO_NAME.dump"

 # Create the repo(which is FSFS by default)
  su - $HTTPD_USER -c "$SVNADMIN create /tmp/$REPO_NAME"
  su - $HTTPD_USER -c "$SVNADMIN load -q /tmp/$REPO_NAME < /tmp/$REPO_NAME.dump"
  REV_AFTER_LOAD=`su - $HTTPD_USER -c "$SVNLOOK youngest /tmp/$REPO_NAME"`

 # Check whether all revisions are loaded properly
  if [ $REV_BEFORE_DUMP -ne $REV_AFTER_LOAD ];
  then
    echo -e "There are missing revisions after dump&load\n"
    su - $HTTPD_USER -c "mv $TEMP_REPO $REPO"
    su - $HTTPD_USER -c "rm -f /tmp/$REPO_NAME.dump"
    exit 2
  fi

 # Move the repo to the desired/original location if everything was fine
  su - $HTTPD_USER -c "mv /tmp/$REPO_NAME $REPO_LOCATION"
  if [ $? -ne 0 ];
  then
    echo -e "Couldn't move the repo to the original location, check for proper permissions\n"
    su - $HTTPD_USER -c "mv $TEMP_REPO $REPO"
    su - $HTTPD_USER -c "rm -rf /tmp/$REPO_NAME /tmp/$REPO_NAME.dump"
    exit 2
  fi

 # Remove the tmp repo and the dump on completion
  su - $HTTPD_USER -c "rm -rf $TEMP_REPO /tmp/$REPO_NAME.dump"
  echo -e "Converted the bdb repo '$REPO_NAME' to fsfs\n"
else 
  echo -e "There's already a dump available for $REPO_NAME,\ 
            check for the repo existence and run the script\n"
fi

