#!/bin/bash

SVN_VERSION_EXPECTED='1.8'
SVN_COMMAND=`which svn`

initialise_variables() {
   OPTIONS_FILE="$RUNTIME_DIR"/scripts/options-to-shell-variables.py
   if [ -e "$OPTIONS_FILE" ]; then
        eval "`$OPTIONS_FILE`"
   else
        echo -e "\nERROR: The file $OPTIONS_FILE does not exist\n\nPlease recreate runtime and try again\n"
        exit 1
   fi

   PUBLISHING_PATH=$DATA_DIR/publish
   BRANDING_PATH=$DATA_DIR/overrides
   CLISERVER_DATA_DIR=$DATA_DIR/cliserver
}

get_total_wc() {
    total=`find $PUBLISHING_PATH -mindepth 1 -maxdepth 1 -type d 2> /dev/null | wc -l`
    return $total
}

do_svn_upgrade_branding() {

   if [ ! -d $BRANDING_PATH ];
   then
      return
   fi

   echo -e "***** Doing SVN upgrade on [$BRANDING_PATH]. *****"
   su $HTTPD_USER -c "$SVN_COMMAND upgrade $BRANDING_PATH" 2> /dev/null
   echo -e "***** SVN upgrade on [$BRANDING_PATH] done. *****"
}


do_svn_upgrade_publish() {
    if [ ! -d $PUBLISHING_PATH ];
    then
       return
    fi

    for opDir in `find $PUBLISHING_PATH -mindepth 1 -maxdepth 1 -type d`
    do
         pubDir=$opDir/www/
         if [ -d $pubDir ];
         then
             echo -e "***** Doing SVN upgrade on [$pubDir]. *****"
             su $HTTPD_USER -c "$SVN_COMMAND upgrade $pubDir" 2> /dev/null
             echo -e "***** SVN upgrade on [$pubDir] done. *****"
         fi
    done
}

do_svn_upgrade_cli() {

    if [ ! -d $CLISERVER_DATA_DIR ];
    then
       return
    fi

    for opDir in `find $CLISERVER_DATA_DIR -maxdepth 3 -type d -iname '.svn'`
    do
         scnDir=`dirname $opDir`
         if [ -d $scnDir ];
         then
             echo -e "***** Doing SVN upgrade on [$scnDir]. *****"
             $SVN_COMMAND upgrade $scnDir 2> /dev/null
             echo -e "***** SVN upgrade on [$scnDir] done. *****"
         fi
    done
}

main() {

   initialise_variables
   get_total_wc
   total_wc=$?

   echo -e "You have $total_wc Working copies in publish to be upgraded."
   echo -e "You have 1 Working copies in branding to be upgraded."

   if [ ! -z $1 ]
   then
      if [ $1 != "-Y" ]
      then
          echo -en "Shall I proceed (Y/N) ? "
          read option
      else
          option='Y'
      fi
   else
          echo -en "Shall I proceed (Y/N) ? "
          read option
   fi

   if [ -z $option ] 
   then
      echo -e "SVN working copy upgrade cancelled"
      exit 1
   elif [ $option != 'Y' ] 
   then
      echo -e "SVN working copy upgrade cancelled"
      exit 1
   fi

   echo -e "INFO: SVN Working copies upgrade started."
   do_svn_upgrade_publish
   do_svn_upgrade_branding
   do_svn_upgrade_cli
   echo -e "INFO: SVN Working copies upgrade has been completed."
}

check_run_permission() {
    if [ `id -u` -ne 0 ]; then
        echo -e "Permission Denied to run this program."
        echo -e "NOTE: You must run this script as root!"
        exit 1
    fi

}

check_svn_version() {
   svnVersion=`$SVN_COMMAND --version | head -n1 | (read one two three four; echo "$three")`
   mmsvnVersion=${svnVersion:0:3}   # Require to match only major:minor version nos.

   if [ $mmsvnVersion != $SVN_VERSION_EXPECTED ]
   then
       echo -e "NOTE: This script expects subversion $SVN_VERSION_EXPECTED.x to have installed."
       echo -e "      But what you have is subversion $svnVersion"
       echo -e "      So please upgrade your subversion to $SVN_VERSION_EXPECTED.x first."
       exit 1
   fi
}

check_svn_version
check_run_permission
main $1
